// This file is part of Eigen, a lightweight C++ template library
// for linear algebra. Eigen itself is part of the KDE project.
//
// Copyright (C) 2006 Benoit Jacob <jacob@math.jussieu.fr>
//
// Eigen is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 or (at your option) any later version.
//
// Eigen is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along
// with Eigen; if not, write to the Free Software Foundation, Inc., 51
// Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. This exception does not invalidate any other reasons why a work
// based on this file might be covered by the GNU General Public License.

#ifndef EIGEN_TESTS_MAIN_H
#define EIGEN_TESTS_MAIN_H

#include <QtTest/QtTest>
#include <iostream>
#include <complex>
#include <stdlib.h>
#include <time.h>

#include "../linearsolver.h"
#include "../regression.h"

using namespace Eigen;

void pickRandom( float & x )
{
    x = 2.0f * rand() / RAND_MAX - 1.0f;
}

void pickRandom( double & x )
{
    x = 2.0 * rand() / RAND_MAX - 1.0;
}

template<typename T> void pickRandom( std::complex<T> & x )
{
    pickRandom( x.real() );
    pickRandom( x.imag() );
}

template< typename T,
          int Size,
          typename VectorType,
          typename BigVecType >
void makeNoisyCohyperplanarPoints( int numPoints,
                                   VectorType * points,
                                   BigVecType * coeffs,
                                   T noiseAmplitude )
{
    int i, j;

    // pick a random hyperplane, store the coefficients of its equation
    coeffs->resize( Size + 1 );
    for( j = 0; j < Size + 1; j++ )
    {
        (*coeffs)(j) = static_cast<T>(0);
        while( Util::abs( (*coeffs)(j) ) < 0.5 ) pickRandom((*coeffs)(j));
    }

    // now pick numPoints random points on this hyperplane
    for( i = 0; i < numPoints; i++ )
    {
        VectorType cur_point(Size);
        do
        {
            // make cur_point a random vector
            for( j = 0; j < Size; j++ )
                pickRandom( cur_point(j) );
            cur_point.normalize();

            // project cur_point onto the hyperplane
            T x = static_cast<T>(0);
            for( j = 0; j < Size; j++ )
                x -= (*coeffs)(j) * cur_point(j);
            cur_point *= (*coeffs)(Size) / x;
        } while( Util::abs(cur_point.norm()) < 0.5
              || Util::abs(cur_point.norm()) > 2.0 );

        points[i] = cur_point;
    }

    // add some noise to these points
    for( i = 0; i < numPoints; i++ )
    {
        for( j = 0; j < Size; j++ )
        {
            T someNoise;
            pickRandom( someNoise );
            points[i](j) += someNoise * noiseAmplitude;
        }
    }
}

template< typename T,
          typename VectorType,
          typename BigVecType,
          typename MatrixType >
void helper_checkComputeFittingHyperplane( int numPoints,
                                           const VectorType * points,
                                           const BigVecType * coeffs,
                                           T allowedError )
{
    int size = points[0].size();
    BigVecType result( size + 1 );
    computeFittingHyperplaneBase< T, VectorType,
                                  BigVecType, MatrixType >
        ( numPoints, points, & result );
    result /= result( size );
    result *= (*coeffs)( size );
    T error = (result - *coeffs).norm() / coeffs->norm();
    QVERIFY( Util::abs(error) < Util::abs(allowedError) );
}

class MainTest : public QObject
{
    Q_OBJECT

public:
    MainTest()
    {
        unsigned int t = (unsigned int) time( NULL );
        qDebug() << "Initializing random number generator with seed"
                 << t << endl;
        srand(t);
    }

private slots:
    void checkLinearSolver();
    void checkLinearRegression();
};

#endif
